C**********************************************************
C     ****     MDO ... MINIMUM DEGREE ORDERING    ****
C**********************************************************
C
      SUBROUTINE MDO(AATPAT,AATPNT,CLIQS,MAXNZL,MAXM,M,NZL,
     X PERM,INVP,DGHEAD,LINKFD,LINKBK,
     X RWLIST,LSTCLQ,MARKER,TEMP,IOLOG)
C
C
C *** PARAMETERS
      INTEGER*4 MAXNZL,MAXM,M,NZL,IOLOG
      INTEGER*4 AATPNT(MAXM+1),CLIQS(MAXNZL)
      INTEGER*4 LSTCLQ(MAXM),MARKER(MAXM),TEMP(MAXM),RWLIST(MAXM)
C
C *** The following arrays can be half-length integer.
      INTEGER*2 DGHEAD(MAXM),LINKFD(MAXM),LINKBK(MAXM)
      INTEGER*2 AATPAT(MAXNZL),PERM(MAXM),INVP(MAXM)
C
C
C *** LOCAL VARIABLES
      INTEGER*4 I,IROW,IR,BESTRW,ELROWS,K,KX,KBEG,KEND,NEXT,PREVS
      INTEGER*4 MINDEG,DEGREE,OLDDEG,NEWDEG,DISCRD,MASSEL
      INTEGER*4 NACTCL,PCLQHD,ICLIQ,IBEG,IEND,LENOFL
      DOUBLE PRECISION A1,A2,FLOPS
      CHARACTER*100 BUFFER
C
C
C *** COMMON ARREAS
C     Markers for linking rows.
      COMMON /ICGRAD/ MSPLIT(100000)
      INTEGER*2       MSPLIT
C
C
C *** PARAMETERS DESCRIPTION
C
C     ON INPUT:
C     AATPAT  Sparsity pattern of  A*Atransp handled as
C             a collection of sparse row vectors (diagonal
C             elements are excluded from the list).
C     AATPNT  Pointers to rows of  A*Atransp.
C     MAXNZL  Maximum number of nonzeros of the Cholesky factor.
C     MAXM    Maximum row dimension of the matrix to be decomposed.
C     M       Dimension of the matrix to be decomposed.
C     IOLOG   Output unit number where log messages are to be written.
C
C     ON OUTPUT:
C     NZL     Number of nonzero entries in Cholesky factor.
C     PERM    Permutation resulting from the minimum degree ordering.
C     INVP    Inverse permutation.
C
C     WORK ARRAYS:
C     CLIQS   Cliques of the pivotal rows (linked lists).
C     DGHEAD  Headers of the forward linked lists of rows (nodes)
C             with the same degree.
C     LINKFD  Forward linked lists of rows with the same degree.
C     LINKBK  Backward linked lists of rows with the same degree.
C     LSTCLQ  A list of headers to different pivotal cliques
C             that are still active i.e. that have not yet been
C             merged with any pivotal row.
C     RWLIST  A list of nonzero positions of a row that is
C             involved in a current step of elimination.
C     MARKER  Array used to mark already reordered rows.
C     TEMP    Temporary array used for merging lists.
C
C
C *** SUBROUTINES CALLED:
C     MYWRT
C
C
C *** PURPOSE:
C     This routine implements the minnimum degree ordering
C     for a symmetric positive definite matrix.
C
C
C *** NOTES:
C     1. This routine follows Duff et al. (1989) description
C        of the minimum degree ordering. 
C     2. This routine assumes that the matrix  A*Atransp is
C        positive definite, i.e., that pivoting in the numerical
C        phase will not be required.
C     3. The cliques discarding mechanism has been implemented.
C     4. Mass elimination technique has been implemented.
C
C
C *** REFERENCES:
C     Duff I.S., Erisman A.M., Reid J.K. (1989). Direct methods
C        for sparse matrices, Clarendon Press, Oxford 1989,
C        chapter 10.
C     Gondzio J. (1993). Implementing Cholesky factorization
C        for interior point methods of linear programming,
C        Optimization 27, pp. 121-140.
C
C
C *** HISTORY:
C     Written by:    Jacek Gondzio, Systems Research Institute,
C                    Polish Academy of Sciences, Newelska 6,
C                    01-447 Warsaw, Poland.
C     Date written:  May 8, 1991
C     Last modified: May 17, 1995
C
C
C
C *** BODY OF (MDO) ***
C
C
C     Initialize for the minimum degree ordering.
C     Zero work arrays.
      DO 20 IROW=1,M
         DGHEAD(IROW)=0
         MARKER(IROW)=0
         TEMP(IROW)=0
   20 CONTINUE
C
C     Set the degree doubly linked lists (recall that diagonal
C     elements are not stored in the sparsity pattern).
      DO 40 IROW=1,M
         DEGREE=AATPNT(IROW+1)-AATPNT(IROW)+1
         IF(MSPLIT(IROW).EQ.1) DEGREE=M
         NEXT=DGHEAD(DEGREE)
         LINKFD(IROW)=NEXT
         DGHEAD(DEGREE)=IROW
         IF(NEXT.GT.0) LINKBK(NEXT)=IROW
         LINKBK(IROW)=-DEGREE
   40 CONTINUE
C
C
C *** DEBUGGING
C     DO 42 IROW=1,M
C     WRITE(BUFFER,41) IROW,DGHEAD(IROW),LINKFD(IROW),LINKBK(IROW)
C  41 FORMAT(1X,'MDO: row',I6,'  hd=',I6,'  fd=',I6,'  bk=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C  42 CONTINUE
C
C     Set the parameters controlling the progress
C     of the minimum degree ordering.
C     NACTCL  is the number of the active pivotal cliques
C             that have not yet been merged.
C     ELROWS  is the number of already eliminated rows + 1.
C     MINDEG  is a current minimum degree found.
C     LENOFL  is a current length of the Cholesky factor.
C     FLOPS   is a cost of the numerical phase of the factorization.
      NACTCL=0
      ELROWS=1
      MINDEG=1
      LENOFL=0
      FLOPS=0.0D0
C
C
C
C     Eliminate all single-element rows (isolated nodes).
C     This elimination does not involve any modification
C     of the degree doubly linked lists.
      IROW=DGHEAD(1)
  100 IF(IROW.EQ.0) GO TO 120
C     WRITE(BUFFER,101) IROW
C 101 FORMAT(1X,'MDO: row',I6,' has degree 1')
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Eliminate row IROW (save its position in a permuted matrix).
      INVP(IROW)=ELROWS
      MARKER(IROW)=1
      ELROWS=ELROWS+1
      IROW=LINKFD(IROW)
      GO TO 100
C
C     Here if all single-element rows have been eliminated.
  120 DGHEAD(1)=0
C
C
C
C
C
C     Main loop begins here.
C     Eliminating rows (nodes) of minimum degree.
  200 IF(ELROWS.GT.M) GO TO 2100
C
C     Look for the row of the minimum degree.
  220 IROW=DGHEAD(MINDEG)
      IF(IROW.GT.0) GO TO 240
      MINDEG=MINDEG+1
      GO TO 220
C
C
C     The row of minimum degree has been found.
  240 DEGREE=MINDEG
      BESTRW=IROW
      MARKER(BESTRW)=1
      TEMP(BESTRW)=1
C
C
C *** DEBUGGING
C     WRITE(BUFFER,241) IROW,MINDEG
C 241 FORMAT(1X,'MDO: row',I6,' has minimum degree',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Remove row BESTRW from the linked list of rows with degree MINDEG.
      NEXT=LINKFD(BESTRW)
      IF(NEXT.GT.0) LINKBK(NEXT)=-MINDEG
      DGHEAD(MINDEG)=NEXT
C
C
C
C     Create the current pivotal clique.
C     First, move the pivot row to the form of a linked list
C     and save its sparsity pattern in TEMP array.
C     Pack the pivotal clique to  RWLIST array.
C     DEGREE is a number of entries of the pivotal clique.
C     PCLQHD is a header to the current pivotal clique.
      KBEG=AATPNT(BESTRW)
      KEND=AATPNT(BESTRW+1)-1
      DEGREE=0
      PCLQHD=0
      DO 300 K=KBEG,KEND
         IR=AATPAT(K)
         IF(MARKER(IR).EQ.1) GO TO 300
         DEGREE=DEGREE+1
         TEMP(IR)=-DEGREE
         RWLIST(DEGREE)=IR
         CLIQS(K)=PCLQHD
         PCLQHD=K
  300 CONTINUE
C
C     Scan all the previously determined pivotal cliques
C     if they involve the pivot. If so, then merge the clique
C     with the current one (and update TEMP array).
C     ICLIQ indicates the clique that is being analysed.
C     IBEG  indicates the first clique to be analysed.
C     IEND  indicates the last clique to be analysed.
      IBEG=1
  320 IEND=NACTCL
      IF(IBEG.GT.IEND) GO TO 480
      DO 400 ICLIQ=IBEG,IEND
         K=LSTCLQ(ICLIQ)
  360    IF(K.EQ.0) GO TO 400
         IR=AATPAT(K)
         IF(IR.EQ.BESTRW) GO TO 420
         K=CLIQS(K)
         GO TO 360
  400 CONTINUE
C
C
C     Here if neither of pivotal cliques involves the pivot.
C     The current pivotal clique is determined.
      GO TO 480
C
C
C     Here if a clique that involves the pivot has been found.
C     Merge it with a current pivotal clique.
C     ICLIQ indicates the clique to be merged.
  420 K=LSTCLQ(ICLIQ)
  440 NEXT=CLIQS(K)
      IR=AATPAT(K)
      IF(TEMP(IR).NE.0) GO TO 460
C
C     Add the element to the current pivotal clique.
      DEGREE=DEGREE+1
      TEMP(IR)=-DEGREE
      RWLIST(DEGREE)=IR
      CLIQS(K)=PCLQHD
      PCLQHD=K
  460 K=NEXT
      IF(K.NE.0) GO TO 440
C
C     Remove the merged clique from the list of active ones and
C     return to the scanning for other cliques that involve the pivot.
      LSTCLQ(ICLIQ)=LSTCLQ(NACTCL)
      IBEG=ICLIQ
      NACTCL=NACTCL-1
      GO TO 320
C
C
C
C     Here if the current pivotal clique is determined.
C     It is handled in the form of a linked list (that starts
C     from PCLQHD). The sparsity pattern of this new pivotal
C     clique is also stored in TEMP array. Add the clique
C     to the list of active ones.
  480 NACTCL=NACTCL+1
      LSTCLQ(NACTCL)=PCLQHD
C
C
C     Update  LENOFL and  FLOPS.
      LENOFL=LENOFL+DEGREE
      FLOPS=FLOPS+DBLE(DEGREE)*DBLE(DEGREE)
C
C
C *** DEBUGGING
C     WRITE(BUFFER,481) BESTRW,ELROWS,DEGREE+1
C 481 FORMAT(1X,'MDO: row',I6,' becomes ',I6,' degree=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C
C     Recalculate the degrees.
C     Analyse all rows involved in this pivotal step (included
C     into the current pivotal clique) if their degrees change.
      K=PCLQHD
  500 IF(K.EQ.0) GO TO 860
      IROW=AATPAT(K)
C
C     Do not recalculate degrees of split dense columns (they 
C     remain always at the last position in the reordering).
      IF(MSPLIT(IROW).EQ.1) GO TO 850
C
C     OLDDEG is an old degree of row (node) IROW.
C     NEWDEG is a new degree of row (node) IROW.
C     Recall that pivot element contributes to a degree.
      KX=IROW
  520 KX=LINKBK(KX)
      IF(KX.GT.0) GO TO 520
      OLDDEG=-KX
      NEWDEG=DEGREE
C
C
C
C     Scan all the previously determined pivotal cliques
C     if they involve row  IROW. If so, then merge the clique
C     with the current one (and update TEMP array).
C     ICLIQ indicates the clique that is being analysed.
C     IBEG  indicates the first clique to be analysed.
C     IEND  indicates the last clique to be analysed.
      IBEG=1
      IEND=NACTCL-1
  540 IF(IBEG.GT.IEND) GO TO 780
      DO 580 ICLIQ=IBEG,IEND
         KX=LSTCLQ(ICLIQ)
  560    IF(KX.EQ.0) GO TO 580
         IR=AATPAT(KX)
         IF(IR.EQ.IROW) GO TO 600
         KX=CLIQS(KX)
         GO TO 560
  580 CONTINUE
C
C
C     Here if neither of pivotal cliques involves row IROW.
C     Row  IROW sparsity pattern is determined.
      GO TO 780
C
C
C     Here if the clique that involves row  IROW has been found.
C     Merge it with already determined part of row  IROW
C     sparsity pattern.
C     ICLIQ  indicates the clique to be merged.
C     DISCRD indicates whether the clique can be discarded.
  600 DISCRD=0
      KX=LSTCLQ(ICLIQ)
  700 NEXT=CLIQS(KX)
      IR=AATPAT(KX)
      IF(TEMP(IR)) 760,720,740
C
C     Add the element to the sparsity pattern of row  IROW.
  720 TEMP(IR)=1
      NEWDEG=NEWDEG+1
      RWLIST(NEWDEG)=IR
C     WRITE(BUFFER,721) IR
C 721 FORMAT(1X,'MDO: row',I6,' is added to the pattern of row IROW')
C     CALL MYWRT(IOLOG,BUFFER)
  740 DISCRD=1
  760 KX=NEXT
      IF(KX.NE.0) GO TO 700
C
C     Here if merging is completed.
C     Discard the clique if it is a subset of the current pivotal one.
      IF(DISCRD.EQ.0) LSTCLQ(ICLIQ)=0
C
C
C     Return to the scanning for other cliques that involve row  IROW.
      IBEG=ICLIQ+1
      GO TO 540
C
C
C
C     Here if all the previous cliques that involve row  IROW
C     have been merged with the pivotal clique.
C     The resulting sparsity pattern of row  IROW is handled
C     in  RWLIST array.
C     Merge the original row  IROW with the above list.
  780 KBEG=AATPNT(IROW)
      KEND=AATPNT(IROW+1)-1
      DO 800 KX=KBEG,KEND
         IR=AATPAT(KX)
         IF(MARKER(IR).EQ.1) GO TO 800
         IF(TEMP(IR).NE.0) GO TO 800
C
C     Add the element to the sparsity pattern of row  IROW.
         TEMP(IR)=1
         NEWDEG=NEWDEG+1
         RWLIST(NEWDEG)=IR
  800 CONTINUE
C
C
C     Here if the new degree of row  IROW is fully determined.
C     Update the degree doubly linked lists.
C     Remove row IROW from the linked list of rows with degree OLDDEG.
      IF(NEWDEG.EQ.OLDDEG) GO TO 820
      NEXT=LINKFD(IROW)
      PREVS=LINKBK(IROW)
      IF(NEXT.GT.0) LINKBK(NEXT)=PREVS
      IF(PREVS.LT.0) THEN
         DGHEAD(OLDDEG)=NEXT
      ELSE
         LINKFD(PREVS)=NEXT
      ENDIF
C
C     Add row IROW to the linked list of rows with degree NEWDEG.
      IF(NEWDEG.EQ.0) GO TO 820
      NEXT=DGHEAD(NEWDEG)
      DGHEAD(NEWDEG)=IROW
      LINKBK(IROW)=-NEWDEG
      LINKFD(IROW)=NEXT
      IF(NEXT.GT.0) LINKBK(NEXT)=IROW
C
C
C     Here if linked lists are updated.
C     Restore  RWLIST and  TEMP arrays to the form that only sparsity
C     pattern of the current pivotal clique is taken into account.
  820 DO 840 KX=DEGREE+1,NEWDEG
         IR=RWLIST(KX)
         TEMP(IR)=0
  840 CONTINUE
C
C
C     Continue recalculating the degrees of rows involved
C     in a current step of elimination.
  850 K=CLIQS(K)
      GO TO 500
C
C
C
C     Here if degrees are recalculated.
C     Eliminate row BESTRW (save its position in a permuted matrix).
  860 INVP(BESTRW)=ELROWS
      ELROWS=ELROWS+1
C
C
C
C     Check for a mass elimination case.
C     If the current elimination step have created at least one
C     row with degree equal to  MINDEG-1, then the mass elimination
C     can be done.
      MASSEL=0
      IF(MINDEG.EQ.1) GO TO 910
      IF(DGHEAD(MINDEG-1).EQ.0) GO TO 910
C
C
C     Here to perform mass elimination.
C
C
C *** DEBUGGING
C     WRITE(BUFFER,861)
C 861 FORMAT(1X,'MDO: mass elimination starts.')
C     CALL MYWRT(IOLOG,BUFFER)
C
C     Eliminate all rows with degree equal to  MINDEG-1.
C     Count the eliminated rows.
      DEGREE=MINDEG-1
      IROW=DGHEAD(DEGREE)
  870 IF(IROW.EQ.0) GO TO 880
C
C *** DEBUGGING
C     WRITE(BUFFER,871) IROW,DEGREE,ELROWS
C 871 FORMAT(1X,'MDO: row',I6,' (degree',I6,')  becomes row',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C
C
C     Eliminate row IROW (save its position in a permuted matrix).
      MASSEL=MASSEL+1
      INVP(IROW)=ELROWS
      MARKER(IROW)=1
C
C     Update  LENOFL and  FLOPS.
      DEGREE=DEGREE-1
      LENOFL=LENOFL+DEGREE
      FLOPS=FLOPS+DBLE(DEGREE)*DBLE(DEGREE)
C
      ELROWS=ELROWS+1
      IROW=LINKFD(IROW)
      GO TO 870
C
C
C     End of mass elimination.
  880 DGHEAD(MINDEG-1)=0
C
C
C     Once again update the degrees of rows from the pivotal clique.
C     Take account of the degree changes caused be the mass elimination.
      DO 900 K=1,MINDEG-1
         IROW=RWLIST(K)
         IF(MARKER(IROW).EQ.1) GO TO 900
C
C     OLDDEG is an old degree of row (node) IROW.
C     NEWDEG is a new degree of row (node) IROW.
C     Recall that pivot element contributes to a degree.
         KX=IROW
  890    KX=LINKBK(KX)
         IF(KX.GT.0) GO TO 890
         OLDDEG=-KX
         NEWDEG=OLDDEG-MASSEL
C
C     Update the degree doubly linked lists.
C     Remove row IROW from the linked list of rows with degree OLDDEG.
C        WRITE(BUFFER,891) IROW,OLDDEG,NEWDEG
C 891    FORMAT(1X,'MDO: row ',I6,'  olddeg=',I6,'  newdeg=',I6)
C        CALL MYWRT(IOLOG,BUFFER)
         IF(NEWDEG.EQ.OLDDEG) GO TO 900
         NEXT=LINKFD(IROW)
         PREVS=LINKBK(IROW)
         IF(NEXT.GT.0) LINKBK(NEXT)=PREVS
         IF(PREVS.LT.0) THEN
            DGHEAD(OLDDEG)=NEXT
         ELSE
            LINKFD(PREVS)=NEXT
         ENDIF
C
C     Add row IROW to the linked list of rows with degree NEWDEG.
         IF(NEWDEG.EQ.0) GO TO 900
         NEXT=DGHEAD(NEWDEG)
         DGHEAD(NEWDEG)=IROW
         LINKBK(IROW)=-NEWDEG
         LINKFD(IROW)=NEXT
         IF(NEXT.GT.0) LINKBK(NEXT)=IROW
C
  900 CONTINUE
C
C
C *** DEBUGGING
C     DO 902 IROW=1,M
C     WRITE(BUFFER,901) IROW,DGHEAD(IROW),LINKFD(IROW),LINKBK(IROW)
C 901 FORMAT(1X,'MDO:   row',I6,'  hd=',I6,'  fd=',I6,'  bk=',I6)
C     CALL MYWRT(IOLOG,BUFFER)
C 902 CONTINUE
C
C
C
C     Restore zero value of TEMP array
C     (it still contains the current pivotal clique).
  910 DO 920 KX=1,MINDEG-1
         IR=RWLIST(KX)
         IF(MARKER(IR).NE.0) GO TO 920
         TEMP(IR)=0
  920 CONTINUE
C
C
C
C
C     Remove all the discarded cliques from the list of active ones.
C     ICLIQ indicates the clique that is being analysed.
C     IBEG  indicates the first clique to be analysed.
C     IEND  indicates the last clique to be analysed.
      IBEG=1
  940 IEND=NACTCL
      IF(IBEG.GT.IEND) GO TO 1000
      DO 960 ICLIQ=IBEG,IEND
         IF(LSTCLQ(ICLIQ).EQ.0) GO TO 980
  960 CONTINUE
      GO TO 1000
C
C     Compress the list.
  980 LSTCLQ(ICLIQ)=LSTCLQ(NACTCL)
      IBEG=ICLIQ
      NACTCL=NACTCL-1
      GO TO 940
C
C
C
C
C     End of main loop.
C     Row of the minimum degree has been eliminated.
 1000 MINDEG=MINDEG-MASSEL-1
      IF(MINDEG.LT.1) MINDEG=1
      GO TO 200
C
C
C
C
C
C     Minimum degree heuristic is completed.
C     Set the permutation vector.
 2100 DO 2200 I=1,M
         K=INVP(I)
         PERM(K)=I
 2200 CONTINUE
C
C
C
C
C     Write problem statistics.
      K=(AATPNT(M+1)-1)/2
      KX=LENOFL
      NZL=KX
      A1=LENOFL*200.0
      A2=M*M-M
      IF(M.GT.1) THEN
         A1=A1/A2
      ELSE
         A1=0.0
      ENDIF
      WRITE(BUFFER,2201) KX,A1
 2201 FORMAT(1X,'MDO:    Matrix  L will have ',I13,
     X  ' subdiagonal elts (density=',F5.1,'%).')
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,2202) KX-K
 2202 FORMAT(1X,'        Fill-in             ',I13)
      CALL MYWRT(IOLOG,BUFFER)
      WRITE(BUFFER,2203) FLOPS
 2203 FORMAT(1X,'        Decomposition flops',1PD14.6)
      CALL MYWRT(IOLOG,BUFFER)
C
C
      RETURN
C
C
C
C *** LAST CARD OF (MDO) ***
      END
